/**
 *  Licensed to the Apache Software Foundation (ASF) under one or more
 *  contributor license agreements.  See the NOTICE file distributed with
 *  this work for additional information regarding copyright ownership.
 *  The ASF licenses this file to You under the Apache License, Version 2.0
 *  (the "License"); you may not use this file except in compliance with
 *  the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.apache.geronimo.naming.java;

import java.util.Collections;

import javax.naming.Context;
import javax.naming.Name;
import javax.naming.NameNotFoundException;
import javax.naming.NamingException;

import org.apache.xbean.naming.context.ImmutableContext;

/**
 * The root context for the java: namespace.
 * Automatically handles switching the "java:comp" sub-context to the
 * appropriate one for the current thread.
 *
 * @version $Rev: 901935 $ $Date: 2010-01-21 19:10:53 -0500 (Thu, 21 Jan 2010) $
 */
public class RootContext extends ImmutableContext {
    private static InheritableThreadLocal<Context> compContext = new InheritableThreadLocal<Context>();

    public RootContext() throws NamingException {
        super(Collections.<String, Object>emptyMap());
    }

    @Override
    public Object lookup(String name) throws NamingException {
        if (name.startsWith("java:")) {
            Context compCtx = compContext.get();
            if (compCtx == null) {
                // the component context was not set for this thread
                throw new NameNotFoundException("No thread context set, looking up: " + name);
            }
            if (name.length() == 5) {
                return compCtx;
            }
            name = name.charAt(5) == '/'? name.substring(6): name.substring(5);
            return compCtx.lookup(name);


//            if ("comp".equals(name)) {
//                return compCtx;
//            } else if (name.startsWith("comp/")) {
//                return compCtx.lookup(name.substring(5));
//            } else if ("/comp".equals(name)) {
//                return compCtx;
//            } else if (name.startsWith("/comp/")) {
//                return compCtx.lookup(name.substring(6));
//            } else {
//                throw new NameNotFoundException("Unrecognized name, does not start with expected 'comp': " + name);
//            }
        }
        return super.lookup(name);
    }

    @Override
    public Object lookup(Name name) throws NamingException {
        return lookup(name.toString());
    }

    /**
     * Set the component context for the current thread. This will be returned
     * for all lookups of "java:comp"
     * @param ctx the current components context
     */
    public static void setComponentContext(Context ctx) {
        compContext.set(ctx);
    }

    /**
     * Get the component context for the current thread.
     * @return the current components context
     */
    public static Context getComponentContext() {
        return compContext.get();
    }
}
